#!/usr/bin/env python
# -*- coding: utf-8 -*-
#.--. .-. ... .... -. - ... .-.-.- .. -.

"""
metaRNA finds potential target sites for the microRNAs in genomic sequences.
To use metaRNA use explicit imports.

Example:
>>> from metarna.target_scan import scan, free_energy
>>> scan(**args)
"""
